/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "liberty.h"
#include "patterns.h"

#define MAX_MOVES 2
#define MAX_LUNCHES 5

static int mouse_lunchi[MAX_LUNCHES];
static int mouse_lunchj[MAX_LUNCHES];

struct move_data
{
  int i;            /* move        */
  int j;            /* coordinates */
  int value;        /* value */
  const char *name; /* name of the pattern suggesting the move */
  int same_dragon;  /* whether the move extends the dragon or not */
};

static int mouse_shapes(struct move_data moves[MAX_MOVES], int color,
			char goal[MAX_BOARD][MAX_BOARD], struct pattern *type);
static void mouse_shapes_callback(int m, int n, int color,
				  struct pattern *pattern, int ll, void *data);
static int mouse_determine_life(char goal[MAX_BOARD][MAX_BOARD], 
				struct eye_data eye[MAX_BOARD][MAX_BOARD],
				int color, int *hi, int *hj);
static void mouse_mark(int i, int j, int color, char mx[MAX_BOARD][MAX_BOARD], char mark);


/* Returns true if a move can be found to attack the dragon
 * at (m,n), in which case (*ui,*uj) is the recommended move.
 * (*ui, *uj) can be null pointers if the result is not needed.
 *
 * The array goal marks the extent of the dragon. This must
 * be maintained during reading.
 */

int
mouse_attack (int m, int n, int *ui, int *uj, char goal[MAX_BOARD][MAX_BOARD])
{
  struct move_data moves[MAX_MOVES];
  char newgoal[MAX_BOARD][MAX_BOARD];
  int i, j, k;
  int hi, hj;
  int true_genus;

  if (stackp==0)
    for (i=0; i<board_size; i++)
      for (j=0; j<board_size; j++) {
	if ((dragon[i][j].origini==dragon[m][n].origini)
	    && (dragon[i][j].originj==dragon[m][n].originj))
	  goal[i][j]=1;
	else
	  goal[i][j]=0;
      }
  
  if ((stackp <= depth)
      && (mouse_shapes(moves, OTHER_COLOR(p[m][n]), goal, apat))) {
    /* A move of value 100 is considered a win */
    if (moves[0].value >= 100) {
      if (ui) *ui = moves[0].i;
      if (uj) *uj = moves[0].j;
      return 1;
    }
    /* A move of value 99 is considered a forced move. No other move need
     * be considered. If there are two of these on the board, we lose.
     */
    if ((moves[0].value == 99)
	&& (moves[0].i != -1)) {
	if ((moves[1].value == 99)
	    && (moves[1].i != -1))
	  return 0;
	moves[1].i = -1;
    }
    for (k=0; k<MAX_MOVES; k++) {
      memcpy(newgoal, goal, MAX_BOARD*MAX_BOARD);

      if ((moves[k].i != -1)
	  && trymove(moves[k].i, moves[k].j, OTHER_COLOR(p[m][n]),
		     moves[k].name, m, n)) {
	if ((p[m][n] == EMPTY)
	    || !mouse_defend(m, n, NULL, NULL, newgoal)) {
	  popgo();
	  if (ui) *ui = moves[k].i;
	  if (uj) *uj = moves[k].j;
	  return 1;
	}
	popgo();
      }
    }
  }
  else {
    int save_sgf_dump=sgf_dump;
    int save_count_variations=count_variations;
    sgf_dump=0;
    count_variations=0;
    if (p[m][n]==BLACK)
      true_genus=mouse_determine_life(goal, black_eye, BLACK, &hi, &hj);
    else 
      true_genus=mouse_determine_life(goal, white_eye, WHITE, &hi, &hj);
    sgf_dump=save_sgf_dump;
    count_variations = save_count_variations;
    /* It is now the attacker's move so we win if he has 1.5 eyes. */
    if (true_genus < 4) {
      TRACE("%oVariation %d: DEAD\n", count_variations-1);
      if (true_genus == 3) {
	if (ui) *ui = hi;
	if (uj) *uj = hj;
      }
      else {
	if (ui) *ui = -1;
	if (uj) *uj = -1;
      }
      return 1;
    }
    else 
      TRACE("%oVariation %d: ALIVE\n", count_variations-1);
  }
  return 0;
}
      

/* Returns true if a move can be found to defend the dragon
 * at (m,n), in which case (*ui,*uj) is the recommended move.
 * (*ui, *uj) can be null pointers if the result is not needed.
 *
 * The array goal marks the extent of the dragon. This must
 * be maintained during reading.  */

int
mouse_defend (int m, int n, int *ui, int *uj, char goal[MAX_BOARD][MAX_BOARD])
{
  struct move_data moves[MAX_MOVES];
  char newgoal[MAX_BOARD][MAX_BOARD];
  int i, j, k;
  int hi, hj;
  int true_genus;
  
  if (stackp==0)
    for (i=0; i<board_size; i++)
      for (j=0; j<board_size; j++) {
	if ((dragon[i][j].origini==dragon[m][n].origini)
	    && (dragon[i][j].originj==dragon[m][n].originj))
	  goal[i][j]=1;
	else
	  goal[i][j]=0;
      }
  
  if ((stackp <= depth)
      && (mouse_shapes(moves, p[m][n], goal, dpat))) {
    /* A move of value 100 is considered a win */
    if (moves[0].value >= 100) {
      if (ui) *ui = moves[0].i;
      if (uj) *uj = moves[0].j;
      TRACE("%oVariation %d: ALIVE (Winning dpat)\n", count_variations-1);
      return 1;
    }
    /* A move of value 99 is considered a forced move. No other move need
     * be considered. If there are two of these on the board, we lose.
     */
    if ((moves[0].value == 99)
	&& (moves[0].i != -1)) {
	if ((moves[1].value == 99)
	    && (moves[1].i != -1))
	  return 0;
	moves[1].i = -1;
    }
    for (k=0; k<MAX_MOVES; k++) {
      int i = moves[k].i;
      int j = moves[k].j;
      if ((i != -1)
	  && trymove(i, j, p[m][n], moves[k].name, m, n)) {
	memcpy(newgoal, goal, MAX_BOARD*MAX_BOARD);
	/* Add the stone just played to the goal dragon, unless the
         * pattern explicitly asked for not doing this.
	 */
	if (moves[k].same_dragon)
	  mouse_mark(i, j, p[m][n], newgoal, 1);
	if (!mouse_attack(m, n, NULL, NULL, newgoal)) {
	  popgo();
	  if (ui) *ui = i;
	  if (uj) *uj = j;
	  return 1;
	}
      popgo();
      }
    }
  }
  else {
    int save_sgf_dump=sgf_dump;
    int save_count_variations=count_variations;
    sgf_dump=0;
    count_variations=0;
    if (p[m][n]==BLACK)
      true_genus=mouse_determine_life(goal, black_eye, BLACK, &hi, &hj);
    else 
      true_genus=mouse_determine_life(goal, white_eye, WHITE, &hi, &hj);
    sgf_dump=save_sgf_dump;
    count_variations = save_count_variations;
    /* It is now the defender's move so we win if we has 1.5 eyes. */
    if (true_genus > 2) {
      TRACE("%oVariation %d: ALIVE\n", count_variations-1);
      if (true_genus == 3) {
	if (ui) *ui = hi;
	if (uj) *uj = hj;
      }
      else {
	if (ui) *ui = -1;
	if (uj) *uj = -1;
      }
      return 1;
    }
    else
      TRACE("%oVariation %d: DEAD\n", count_variations-1);	
  }
  return 0;
}


/* The support of goal is a string for which no defensive move is
 * found. If the string is BLACK then eye is black_eye, and if
 * the string is WHITE then eye is white_eye. The function runs make_domains
 * and computes the number of eyes. Returns twice the number of half eyes.
 */

static int
mouse_determine_life(char goal[MAX_BOARD][MAX_BOARD],
		     struct eye_data eye[MAX_BOARD][MAX_BOARD],
		     int color, int *hi, int *hj)
{
  int mw[MAX_BOARD][MAX_BOARD];
  int true_genus = 0;
  int max, min;
  int attacki, attackj;
  int m, n;

  memset(mw, 0, sizeof(mw));
  mouse_find_lunches(goal, color);
  make_domains();

  /* The eyespaces we want to evaluate are the one which
   * are adjacent to the dragon (whose stones comprise the
   * support of goal) which are not GRAY_BORDERED. These
   * are the eyespaces of the dragon. Now we find their
   * origins.
   */

  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      if (p[m][n] && goal[m][n]) {
	if ((m>0) && (p[m-1][n] == EMPTY) 
	    && eye[m-1][n].color != GRAY_BORDER
	    && eye[m-1][n].origini != -1)
	  mw[eye[m-1][n].origini][eye[m-1][n].originj]=1;
	if ((m<board_size-1) && (p[m+1][n] == EMPTY) 
	    && eye[m+1][n].color != GRAY_BORDER
	    && eye[m+1][n].origini != -1)
	  mw[eye[m+1][n].origini][eye[m+1][n].originj]=1;
	if ((n>0) && (p[m][n-1] == EMPTY) 
	    && eye[m][n-1].color != GRAY_BORDER
	    && eye[m][n-1].origini != -1)
	  mw[eye[m][n-1].origini][eye[m][n-1].originj]=1;
	if ((n<board_size-1) && (p[m][n+1] == EMPTY) 
	    && eye[m][n+1].color != GRAY_BORDER
	    && eye[m][n+1].origini != -1)
	  mw[eye[m][n+1].origini][eye[m][n+1].originj]=1;
      }

  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      half_eye[m][n].type = 0;

  /* Find topological half eyes and false eyes by analyzing the
   * diagonal intersections, as described in the Texinfo
   * documentation (Eyes/Eye Topology).
   */

  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      int sum;
      int ci = -1;
      int cj = -1;

      if ((p[m][n]==EMPTY)
	  && (eye[m][n].origini != -1)
	  && (mw[eye[m][n].origini][eye[m][n].originj])
	  && !eye[m][n].marginal
	  && (eye[m][n].neighbors <= 1)) {
	sum = evaluate_diagonal_intersection(m+1, n+1, color, &ci, &cj) 
	  + evaluate_diagonal_intersection(m+1, n-1, color, &ci, &cj) 
	  + evaluate_diagonal_intersection(m-1, n+1, color, &ci, &cj) 
	  + evaluate_diagonal_intersection(m-1, n-1, color, &ci, &cj);
	
	if (sum >= 4) {
	  half_eye[m][n].type = FALSE_EYE;
	  if ((eye[m][n].esize == 1) || legal(m, n, OTHER_COLOR(color)))
	    add_half_eye(m, n, eye);
	}
	else if (sum == 3) {
	  half_eye[m][n].type = HALF_EYE;
	  half_eye[m][n].ki = ci;
	  half_eye[m][n].kj = cj;
	  add_half_eye(m, n, eye);
	}
      }
    }

  /* This test must be conditioned on (m,n) being its own origin,
   * because some origins get moved during the topological eye
   * code.
   */

  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) 
      if ((mw[m][n]) 
	  && (eye[m][n].origini==m)
	  && (eye[m][n].originj==n)) {
	compute_eyes(m, n, &max, &min, &attacki, &attackj, eye);
	true_genus += (max+min);
	if (max != min) {
	  if (hi) *hi=attacki;
	  if (hj) *hj=attackj;
	}
	/* found two eyes---look no further */
	if (true_genus >= 4)
	  return 4;
      }
  return true_genus;
}

	
/* Generates up to MAX_MOVES moves, attempting to attack or defend the
 * dragon at (goali, goalj). The found moves are put in moves, an
 * array of move_data structs. The entries in the array are sorted by
 * value with moves[0] having highest priority. When no more moves are
 * available this is indicated by value and coordinates in the array
 * being -1.
 *
 * Returns 1 if at least one move is found, or 0 if no move is found.
 */

static int 
mouse_shapes(struct move_data moves[MAX_MOVES], int color,
	     char goal[MAX_BOARD][MAX_BOARD], struct pattern *type)
{
  int m, n; /* iterate over all board posns */
  int k;
  int save_sgf_dump=sgf_dump;
  int save_count_variations=count_variations;
  sgf_dump=0;
  count_variations=0;

  for (k=0; k<MAX_MOVES; k++) {
    moves[k].i = -1;
    moves[k].j = -1;
    moves[k].value = -1;
    moves[k].name = NULL;
    moves[k].same_dragon = 1;
  }

  compile_for_match();
  for (m = 0; m < board_size; m++)
    for (n = 0; n < board_size; n++)
      if (p[m][n])
	matchpat(m, n, mouse_shapes_callback, color, type, moves, goal);

  sgf_dump=save_sgf_dump;
  count_variations=save_count_variations;

  if (moves[0].value > 0)
    return 1;
  else
    return 0;
}  



/* This function takes an array of already found moves (passed as
 * 'data') and looks for moves to replace these. Only moves near
 * (goali, goalj) are considered.
 */

static void
mouse_shapes_callback(int m, int n, int color, struct pattern *pattern,
		      int ll, void *data)
{
  int ti, tj, tval;  /* trial move and its value */
  struct move_data *moves = data; /* considered moves passed as data */
  int k;
  
  /* Pick up the location of the move */
  TRANSFORM(pattern->movei, pattern->movej, &ti, &tj, ll);
  ti += m;
  tj += n;

  /* For sacrifice patterns, the survival of the stone to be played is
   * not checked (but it still needs to be legal). Otherwise we
   * discard moves which can be captured. 
   */
  if (! (pattern->class & CLASS_s)) {
    if (!safe_move(ti, tj, color)) {
      TRACE("  move at %m wasn't safe, discarded\n", ti, tj);
      return;
    }
  }
  else {
    if (! legal(ti, tj, color)) {
      TRACE("  move at %m wasn't legal, discarded\n", ti, tj);
      return;
    }
  }
  
  /* For class n patterns, the pattern is contingent on an opponent
   * move at * not being captured.
   */
  if (pattern->class & CLASS_n) {
    if (!safe_move(ti, tj, OTHER_COLOR(color))) {
      TRACE("  opponent can't play safely at %m, move discarded\n", ti, tj);
      return;
    }
  }
  
  /* If the pattern has a constraint, call the autohelper to see
   * if the pattern must be rejected.
   */
  if (pattern->autohelper_flag & HAVE_CONSTRAINT) {
    if (!pattern->autohelper(pattern, ll, ti, tj, color, 0))
      return;
  }

  /* and work out the value of this move */
  if (pattern->helper) {
    /* ask helper function to consider the move */
    DEBUG(DEBUG_HELPER, "  asking helper to consider '%s'+%d at %m\n", 
	  pattern->name, ll, ti,tj);
    tval = pattern->helper(pattern, ll, ti, tj, color);
    
    if (tval > 0) {
      TRACE("helper likes pattern '%s' value %d at %m\n",
	    pattern->name, tval, ti, tj);
    }
    else {
      DEBUG(DEBUG_HELPER,"  helper does not like pattern '%s' at %m\n", 
	    pattern->name, ti, tj);
      return;  /* pattern matcher does not like it */
    }
  }
  else { /* no helper */
    tval = pattern->value;
  }

  /* having made it here, we have made it through all the extra checks */

  TRACE("Pattern %s found at %m with value %d\n", pattern->name, ti, tj, tval);

  /* Add the new move to the list of already found moves, if the value
   * is sufficently large. We keep the list sorted.
   *
   * First we must see if this move already is in the list.
   */
  for (k=0; k<MAX_MOVES; k++) {
    if (moves[k].value == -1)
      break;
    if (moves[k].i == ti && moves[k].j == tj)
      break;
  }

  /* Did we already have this move in the list with a higher value? */
  if (k < MAX_MOVES && moves[k].value >= tval)
    return;

  /* Insert the move at the right place in the list and adjust other
   * entries as needed.
   */
  for (; k>=0; k--) {
    if (k == 0 || tval <= moves[k-1].value) {
      /* Can't get higher. Insert the move below this point and quit
       * looping.
       */
      if (k < MAX_MOVES) {
	moves[k].i = ti;
	moves[k].j = tj;
	moves[k].value = tval;
	moves[k].name = pattern->name;
	/* If B class pattern, this move shouldn't be added to the
         * dragon under consideration.
	 */
	moves[k].same_dragon = ((pattern->class & CLASS_B) == 0);
      }
      break;
    }
    /* Shuffle the passed move one step downwards. */
    if (k < MAX_MOVES)
      moves[k] = moves[k-1]; /* struct copy */
  }

  /* Assert that the list contains unique moves. */
  if (1) {
    int l;
    for (k=0; k<MAX_MOVES; k++)
      for (l=k+1; l<MAX_MOVES; l++)
	assert ((moves[k].i == -1) || (moves[k].i != moves[l].i) || (moves[k].j != moves[l].j));
  }
}


/* Marks all strings of (color) adjacent to (i,j). Also marks (i,j) if p[i][j]==color.
*/

static void
mouse_mark(int i, int j, int color, char mx[MAX_BOARD][MAX_BOARD], char mark)
{
  if (p[i][j] == color)
    mx[i][j] = mark;
  if ((i>0) && p[i-1][j]==color && (!mx[i-1][j]))
    mouse_mark(i-1, j, color, mx, mark);
  if ((i<board_size-1) && p[i+1][j]==color && (!mx[i+1][j]))
    mouse_mark(i+1, j, color, mx, mark);
  if ((j>0) && p[i][j-1]==color && (!mx[i][j-1]))
    mouse_mark(i, j-1, color, mx, mark);
  if ((j<board_size-1) && p[i][j+1]==color && (!mx[i][j+1]))
    mouse_mark(i, j+1, color, mx, mark);
}

/* For use in gdb. Lists the goal array. */

void
goaldump(char goal[MAX_BOARD][MAX_BOARD])
{
  int m, n;
  for (m = 0; m < board_size; m++)
    for (n = 0; n < board_size; n++)
      if (goal[m][n])
	gprintf("%o%m ", m, n);
  gprintf("\n");
}


/* Add mouse reasons */

void
mouse_reasons(int color)
{
  int m, n;
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      if ((dragon[m][n].origini == m)
	  && (dragon[m][n].originj == n)
	  && (dragon[m][n].mouse_status == CRITICAL)
	  && (dragon[m][n].mouse_attacki != -1)
	  && (dragon[m][n].mouse_attackj != -1)) {
	if (p[m][n]==color) {
	  add_mouse_defense_move(dragon[m][n].mouse_defendi,
				 dragon[m][n].mouse_defendj, m, n);
	  if (debug & DEBUG_MOUSE)
	    gprintf("mouse: %m defends %m at move %d\n",
		    dragon[m][n].mouse_defendi,
		    dragon[m][n].mouse_defendj, m, n, movenum+1);
	}
	else {
	  add_mouse_attack_move(dragon[m][n].mouse_attacki,
				dragon[m][n].mouse_attackj, m, n);
	  /* FIXME: take this print statement out when the 
	   * mouse code is better trusted */
	  if (debug & DEBUG_MOUSE)
	    gprintf("mouse: %m attackss %m at move %d\n",
		    dragon[m][n].mouse_attacki,
		    dragon[m][n].mouse_attackj, m, n, movenum+1);
	}
      }
}

/* This function attempts to make a list of dead strings
 * which may be relevant to the life of the goal dragon.
 * Such strings are called mouse lunches. They are ignored
 * (treated as invisible) during the running of make_domains.
 */

void
mouse_find_lunches(char goal[MAX_BOARD][MAX_BOARD], int color)
{
  int m, n;
  int lunches=0;
  int prevlunch;
  int li, lj;
  int other=OTHER_COLOR(color);

  for (prevlunch=0; prevlunch<MAX_LUNCHES; prevlunch++) {
    mouse_lunchi[prevlunch] = -1;
    mouse_lunchj[prevlunch] = -1;
  }
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      if ((p[m][n]==color) && goal[m][n]) {
	if ((m > 0) && (p[m-1][n] == other)) {
	  find_origin(m-1, n, &li, &lj);
	  for (prevlunch=0; prevlunch < lunches; prevlunch++)
	    if ((mouse_lunchi[prevlunch] == li)
		&& (mouse_lunchj[prevlunch] == lj))
	      continue;
	  if (attack(li, lj, NULL, NULL)) {
	    mouse_lunchi[lunches] = li;
	    mouse_lunchj[lunches] = lj;
	    lunches++;
	    if (lunches == MAX_LUNCHES)
	      return;
	  }
	}
	if ((m < board_size-1) && (p[m+1][n] == other)) {
	  find_origin(m+1, n, &li, &lj);
	  for (prevlunch=0; prevlunch < lunches; prevlunch++)
	    if ((mouse_lunchi[prevlunch] == li)
		&& (mouse_lunchj[prevlunch] == lj))
	      continue;
	  if (attack(li, lj, NULL, NULL)) {
	    mouse_lunchi[lunches] = li;
	    mouse_lunchj[lunches] = lj;
	    lunches++;
	    if (lunches == MAX_LUNCHES)
	      return;
	  }
	}
	if ((n > 0) && (p[m][n-1] == other)) {
	  find_origin(m, n-1, &li, &lj);
	  for (prevlunch=0; prevlunch < lunches; prevlunch++)
	    if ((mouse_lunchi[prevlunch] == li)
		&& (mouse_lunchj[prevlunch] == lj))
	      continue;

	  if (attack(li, lj, NULL, NULL)) {
	    mouse_lunchi[lunches] = li;
	    mouse_lunchj[lunches] = lj;
	    lunches++;
	    if (lunches == MAX_LUNCHES)
	      return;
	  }
	}
	if ((n > 0) && (p[m][n+1] == other)) {
	  find_origin(m, n+1, &li, &lj);
	  for (prevlunch=0; prevlunch < lunches; prevlunch++)
	    if ((mouse_lunchi[prevlunch] == li)
		&& (mouse_lunchj[prevlunch] == lj))
	      continue;
	  if (attack(li, lj, NULL, NULL)) {
	    mouse_lunchi[lunches] = li;
	    mouse_lunchj[lunches] = lj;
	    lunches++;
	    if (lunches == MAX_LUNCHES)
	      return;
	  }
	}
      }
}

	  
/* True unless (i, j) is EMPTY or occupied by a lunch for the goal dragon.  
 * Used during make_domains (see optics.c: lively macro).
 */

int 
mouse_lively(int i, int j)
{
  int oi, oj, lunch;

  if (p[i][j]==EMPTY)
    return 0;
  find_origin(i, j, &oi, &oj);
  for (lunch=0; lunch<MAX_LUNCHES; lunch++)
    if ((mouse_lunchi[lunch] == oi) && (mouse_lunchj[lunch] == oj))
      return 0;
  return 1;
}





/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
